﻿using System;
using System.IO;
using ETHZ.C0284H.Calculator.Exceptions;
using ETHZ.C0284H.Calculator.Exceptions.Calc;
using ETHZ.C0284H.Calculator.Instruct;
using ETHZ.C0284H.Calculator.Instruct.Visitor;

namespace ETHZ.C0284H.Calculator.UI
{
    /// <summary>
    /// Console view of a calculator.
    /// </summary>
    internal sealed class CalculatorConsole : ICalculatorView
    {
        // Input of the console.
        private readonly TextReader _reader;

        // Output of the console.
        private readonly TextWriter _writer;

        // Calculator as the model in MVC.
        private readonly ReversePolishNotationCalculator _calculator;

        // Should the console quit?

        // Executor to execute the instructs.
        private readonly ConsoleExecutor _executor;

        /// <summary>
        /// Indicates, whether the console should quit.
        /// </summary>
        private bool _exitRequested;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="reader">Console input.</param>
        /// <param name="writer">Console output.</param>
        /// <param name="calculator">Calculator.</param>
        public CalculatorConsole(TextReader reader, TextWriter writer, ReversePolishNotationCalculator calculator)
        {
            if (reader == null)
            {
                throw new ArgumentNullException("reader");
            }

            if (writer == null)
            {
                throw new ArgumentNullException("writer");    
            }

            if (calculator == null)
            {
                throw new ArgumentNullException("calculator");
            }

            _reader = reader;
            _writer = writer;
            _calculator = calculator;
            _executor = new ConsoleExecutor(this, _calculator);

            _exitRequested = false;
        }

        /// <summary>
        /// Event of new Instruct.
        /// </summary>
        public event EventHandler<InstructEventArgs> InstructEvent;

        /// <summary>
        /// Print message to the output.
        /// </summary>
        /// <param name="message"></param>
        public void PrintMessage(String message)
        {
            _writer.WriteLine(message);
        }

        /// <summary>
        /// Accepts and executes user commands in a loop.
        /// </summary>
        public void Run()
        {
            string cmdText = string.Empty;
            while (!_exitRequested)
            {
                PrintPrompt();
                try
                {
                    cmdText = ReadInput();
                }
                catch (IOException e)
                {
                    PrintMessage(e.Message);
                    Quit();
                }

                if (_exitRequested)
                {
                    break;
                }

                try
                {
                    ICalculatorInstruct instruct = ParseInstructFrom(cmdText);
                    ExecuteInstruct(instruct);
                }
                catch (MissingInstructException)
                {
                    PrintMessage(UIResourceManager.Instance.GetLocalizedMessage(UIMessageID.MSG_INVALID_INPUT));
                    PrintMessage("\t" + cmdText);
                }
            }

            PrintMessage(UIResourceManager.Instance.GetLocalizedMessage(UIMessageID.MSG_QUITTING_APP));
        }

        /// <summary>
        /// Signals the console to quit.
        /// </summary>
        public void Quit()
        {
            _exitRequested = true;
        }

        /// <summary>
        /// Get instruct object from a string.
        /// </summary>
        /// <param name="input">a string representation of instruct.</param>
        /// <returns></returns>
        private static ICalculatorInstruct ParseInstructFrom(string input)
        {
            ICalculatorInstruct instruct;
            if ( InstructRepository.TryParseInstruct(input, out instruct)) 
            {
                return instruct;
            }

            throw new MissingInstructException();
        }

        /// <summary>
        /// Entry point to trigger a new instruct event.
        /// </summary>
        /// <param name="instr"></param>
        /// <param name="sender"></param>
        private void OnNewInstruct(object sender, InstructEventArgs instr)
        {
            if (InstructEvent != null)
            {
                InstructEvent(sender, instr);
            }
        }

        /// <summary>
        /// Print the prompt character.
        /// </summary>
        private void PrintPrompt()
        {
            _writer.Write("#");
        }

        private string ReadInput()
        {
            string input = _reader.ReadLine();

            return input != null ? input.Trim() : String.Empty;
        }

        /// <summary>
        /// Executes the instruct on the console side, if possible;
        /// Otherwise, notifies observers.
        /// </summary>
        /// <param name="instruct">The instruct to execute.</param>
        private void ExecuteInstruct(ICalculatorInstruct instruct)
        {
            instruct.Accept(_executor);
            if (!_executor.InstructProcessed)
            {
                try
                {
                    OnNewInstruct(this, new InstructEventArgs(instruct));
                }
                catch (CalculatorException e)
                {      
                    PrintMessage(e.Message);
                }           
            }
        }

    }
}
