﻿using System;
using System.Collections.Concurrent;
using System.ComponentModel;
using System.Reflection;

namespace ETHZ.C0284H.Calculator.Utils
{
    /// <summary>
    /// Helper class for dealing with enums routines.
    /// </summary>
    public static class Enums
    {
        /// <summary>
        /// A concurrent cache for enum-description pair. Note, that cache should be concurrent, as we are accesing it with static methods.
        /// We could implement cache with locks + Dictionary, but concurrent dictionary is faster than handwritten locking.
        /// </summary>
        private static readonly ConcurrentDictionary<Enum, string> s_cache = new ConcurrentDictionary<Enum, string>();

        /// <summary>
        /// Gets the description of enum, by examining the <see cref="DescriptionAttribute"/>. If such attribute is not present, uses the string
        /// representation. Uses caching to speed-up the performance.
        /// </summary>
        /// <param name="value">Enum value.</param>
        public static string GetEnumDescription(Enum value)
        {
            string result;

            if (s_cache.TryGetValue(value, out result))
            {
                return result;
            }

            result = GetValueInternal(value);

            s_cache.TryAdd(value, result);

            return result;
        }
 
        /// <summary>
        /// Helper method, to fetch custom attributes from 
        /// </summary>
        /// <typeparam name="T">A type of custom attribute</typeparam>
        /// <param name="memberInfo">Any member info. Non null</param>
        /// <returns></returns>
        public static T[] GetAttributes<T>(MemberInfo memberInfo) where T : Attribute
        {
            if (memberInfo == null)
            {
                throw  new ArgumentNullException("memberInfo");
            }

            var result = memberInfo.GetCustomAttributes(typeof (T), false);

            if (result as T[] == null)
            {
                return new T[0];
            }

            return (T[]) result;
        }


        private static string GetValueInternal(Enum value)
        {
            string stringEnum = value.ToString();

            //Reflection : requesting meta-information about enum-field
            FieldInfo fi = value.GetType().GetField(stringEnum);

            //Reflection: trying to fetch Description attribute.
            DescriptionAttribute[] attributes = GetAttributes<DescriptionAttribute>(fi);

            if (attributes.Length > 0)
            {
                return attributes[0].Description;
            }

            return stringEnum;
        }
    }
}